@@ -6,11 +6,12 @@ See ``dir(unpythonic)`` and submodule docstrings for more.
 If you have MacroPy installed, see also ``unpythonic.syntax``.
 """
 
-__version__ = '0.12.0'
+__version__ = '0.14.0'
 
 from .amb import *
 from .arity import *
 from .assignonce import *
+from .collections import *
 from .dynassign import *
 from .ec import *
 from .fold import *
@@ -28,6 +29,13 @@ from .lispylet import let as ordered_let, letrec as ordered_letrec, \
                       blet as ordered_blet, bletrec as ordered_bletrec
 
 from .llist import *
+from .mathseq import *
 from .misc import *
 from .seq import *
+from .slicing import *
 from .tco import *
+
+# HACK: break dependency loop
+from .lazyutil import _init_module
+_init_module()
+del _init_module
