@@ -3,12 +3,7 @@
 
 __all__ = ["env"]
 
-from .dynassign import dyn, make_dynvar
-from .misc import namelambda
-
-# Whether to automatically name lambdas saved in an env.
-# Use  with dyn.let(env_namedlambda=True): ...  to activate.
-make_dynvar(env_namedlambda=False)
+from collections.abc import Container, Sized, Iterable, Mapping, MutableMapping
 
 class env:
     """Environment for let-like constructs.
@@ -48,17 +43,17 @@ class env:
         with env(s="hello", orange="fruit", answer=42) as e:
             ...  # ...code that uses e...
 
-    When the `with` block exits, `e` forgets all its bindings. The `e`
+    When the `with` block exits, ``e`` forgets all its bindings. The ``e``
     instance itself will remain alive due to Python's scoping rules.
     """
     # do not allow bindings that would break functionality.
-    _reserved_names = ("set", "clear", "finalize", "_env", "_allow_more_bindings",
+    _reserved_names = ("set", "clear", "finalize", "_env", "_finalized",
                        "_direct_write", "_reserved_names")
-    _direct_write = ("_env", "_allow_more_bindings")
+    _direct_write = ("_env", "_finalized")
 
     def __init__(self, **bindings):
         self._env = {}
-        self._allow_more_bindings = True  # "let" disables this once env setup done
+        self._finalized = False  # "let" sets this once env setup done
         for name, value in bindings.items():
             setattr(self, name, value)
 
@@ -71,15 +66,13 @@ class env:
             return super().__setattr__(name, value)
         if name in self._reserved_names:
             raise AttributeError("cannot overwrite reserved name '{:s}'; complete list: {}".format(name, self._reserved_names))
-        if not self._allow_more_bindings and name not in self:
+        if self._finalized and name not in self:
             raise AttributeError("name '{:s}' is not defined; adding new bindings to a finalized environment is not allowed".format(name))
         # Block invalid names in subscripting (which redirects here).
         if not name.isidentifier():
             raise ValueError("'{}' is not a valid identifier".format(name))
 #        value = self._wrap(name, value)  # for "e.x << value" rebind syntax.
         self._env[name] = value  # make all other attrs else live inside _env
-        if dyn.env_namedlambda:  # name lambdas saved to an unpythonic.env
-            value = namelambda(value, name)
 
     def __getattr__(self, name):
         # Block invalid names in subscripting (which redirects here).
@@ -90,6 +83,16 @@ class env:
             raise AttributeError("name '{:s}' is not defined".format(name))
         return e[name]
 
+    def __delattr__(self, name):
+        if not name.isidentifier():
+            raise ValueError("'{}' is not a valid identifier".format(name))
+        if self._finalized:
+            raise TypeError("deleting bindings from a finalized environment not allowed; attempted to delete '{:s}'".format(name))
+        e = self._env   # __getattr__ not called if direct attr lookup succeeds, no need for hook.
+        if name not in e:
+            raise AttributeError("name '{:s}' is not defined".format(name))
+        del e[name]
+
     # membership test (in, not in)
     def __contains__(self, k):
         return self._env.__contains__(k)
@@ -99,13 +102,48 @@ class env:
         return self._env.__iter__()
     # no __next__, iterating over dict.
 
+    # Mapping
     def items(self):
         """Like dict.items()."""
         return self._env.items()
+    def keys(self):
+        return self._env.keys()
+    def values(self):
+        return self._env.values()
+    def get(self, k, default=None):
+        return self[k] if k in self else default
+    def __eq__(self, other):
+        return other == self._env
 
     def __len__(self):
         return len(self._env)
 
+    # MutableMapping
+    def pop(self, k, *default):
+        if self._finalized:
+            raise TypeError("deleting bindings from a finalized environment not allowed; attempted to delete '{:s}'".format(k))
+        return self._env.pop(k, *default)
+    def popitem(self):
+        if self._finalized:
+            raise TypeError("deleting bindings from a finalized environment not allowed")
+        return self._env.popitem()
+    def clear(self):
+        if self._finalized:
+            raise TypeError("clearing a finalized environment not allowed")
+        return self._env.clear()
+    def update(self, *mapping, **bindings):
+        if mapping:
+            m = mapping[0]
+            if self._finalized and any(k not in self for k in m):
+                raise AttributeError("adding new bindings to a finalized environment is not allowed")
+        if self._finalized and any(k not in self for k in bindings):
+            raise AttributeError("adding new bindings to a finalized environment is not allowed")
+        return self._env.update(*mapping, **bindings)
+    def setdefault(self, k, *default):
+        if self._finalized and k not in self:
+            raise AttributeError("name '{:s}' is not defined; adding new bindings to a finalized environment is not allowed".format(k))
+        return self._env.setdefault(k, *default)
+
     # subscripting
     def __getitem__(self, k):
         return getattr(self, k)
@@ -113,12 +151,15 @@ class env:
     def __setitem__(self, k, v):
         setattr(self, k, v)
 
+    def __delitem__(self, k):
+        delattr(self, k)
+
     # context manager
     def __enter__(self):
         return self
 
     def __exit__(self, exctype, excvalue, traceback):
-        self.clear()
+        self._env.clear()  # on context exit, clear even if we are a finalized env
 
     # pretty-printing
     def __repr__(self):
@@ -154,19 +195,16 @@ class env:
         self.set(name, value)
         return self
 
-    def clear(self):
-        """Clear the environment, i.e. forget all bindings."""
-        self._env = {}
-
     def finalize(self):
         """Finalize environment.
 
-        This stops the instance from accepting any more new bindings.
+        This stops the instance from accepting any more new bindings,
+        or any deletions of existing bindings.
 
-        Existing bindings can still be overwritten even in a finalized
+        Existing bindings can still be given new values even in a finalized
         environment.
         """
-        self._allow_more_bindings = False
+        self._finalized = True
 
     # For rebind syntax: "e.foo << newval" --> "e.foo.__lshift__(newval)",
     # so foo.__lshift__() must be set up to rebind e.foo.
@@ -202,3 +240,8 @@ class env:
 #                    return cls(obj)  # copy-construct obj without wrapper
 #            assert False, "wrapped value missing in {} {}".format(type(obj), obj)
 #        return _assignonce_wrapper(obj)  # copy-construct obj with wrapper
+
+# register virtual base classes
+for abscls in (Container, Sized, Iterable, Mapping, MutableMapping):
+    abscls.register(env)
+del abscls
