@@ -8,20 +8,23 @@ Requires MacroPy (package ``macropy3`` on PyPI).
 # contain the actual syntax transformers (regular functions that process ASTs)
 # that implement the macros.
 
-# insist, deny, it, f, _, local, block, expr, call_cc are just for passing through
-# to the client code that imports us.
+# insist, deny, it, f, _, force, force1, local, delete, block, expr, call_cc
+# are just for passing through to the client code that imports us.
+from .autoref import autoref as _autoref
 from .curry import curry as _curry
 from .forall import forall as _forall, insist, deny
-from .fupstx import fup as _fup
 from .ifexprs import aif as _aif, it, cond as _cond
 from .lambdatools import multilambda as _multilambda, \
                          namedlambda as _namedlambda, \
-                         quicklambda as _quicklambda, f, _
-from .letdo import do as _do, do0 as _do0, local, \
+                         quicklambda as _quicklambda, f, _, \
+                         envify as _envify
+from .lazify import lazify as _lazify, lazyrec as _lazyrec, force, force1
+from .letdo import do as _do, do0 as _do0, local, delete, \
                    let as _let, letseq as _letseq, letrec as _letrec, \
                    dlet as _dlet, dletseq as _dletseq, dletrec as _dletrec, \
                    blet as _blet, bletseq as _bletseq, bletrec as _bletrec
 from .letsyntax import let_syntax_expr, let_syntax_block, block, expr
+from .nb import nb as _nb
 from .prefix import prefix as _prefix
 from .tailtools import autoreturn as _autoreturn, tco as _tco, \
                        continuations as _continuations, call_cc
@@ -48,6 +51,33 @@ make_dynvar(gen_sym=nogensym)
 
 # -----------------------------------------------------------------------------
 
+@macros.block
+def autoref(tree, args, *, gen_sym, **kw):
+    """Implicitly reference attributes of an object.
+
+    Example::
+
+        e = env(a=1, b=2)
+        c = 3
+        with autoref(e):
+            a
+            b
+            c
+
+    The transformation is::
+
+      x --> o.x if hasattr(o, "x") else x
+
+    The transformation is applied in ``Load`` context only. ``Store`` and ``Del``
+    are not redirected.
+
+    Useful e.g. with the ``.mat`` file loader of SciPy.
+    """
+    with dyn.let(gen_sym=gen_sym):
+        return _autoref(block_body=tree, args=args)
+
+# -----------------------------------------------------------------------------
+
 @macros.expr
 def aif(tree, *, gen_sym, **kw):
     """[syntax, expr] Anaphoric if.
@@ -122,10 +152,9 @@ def curry(tree, **kw):  # technically a list of trees, the body of the with bloc
     builtins in the top-level namespace (as of Python 3.7), but e.g. methods
     of builtin types are not handled.
 
-    In a ``with curry`` block, ``unpythonic.fun.curry`` runs in a special mode
-    that no-ops on uninspectable functions instead of raising ``TypeError``
-    as usual. This special mode is enabled for the *dynamic extent* of the
-    ``with curry`` block.
+    Lexically inside a ``with curry`` block, the auto-curried function calls
+    will skip the curry if the function is uninspectable, instead of raising
+    ``TypeError`` as usual.
 
     Example::
 
@@ -145,8 +174,6 @@ def curry(tree, **kw):  # technically a list of trees, the body of the with bloc
             assert mymap(double, ll(1, 2, 3)) == ll(2, 4, 6)
 
         # The definition was auto-curried, so this works here too.
-        # (Provided add3 contains no calls to uninspectable functions, since
-        #  we are now outside the dynamic extent of the ``with curry`` block.)
         assert add3(1)(2)(3) == 6
     """
     return _curry(block_body=tree)
@@ -275,10 +302,10 @@ def dlet(tree, args, *, gen_sym, **kw):
     lexical scope of the ``def``.
     """
     with dyn.let(gen_sym=gen_sym):
-        return _dlet(bindings=args, fdef=tree)
+        return _destructure_and_apply_let(tree, args, _dlet)
 
 @macros.decorator
-def dletseq(tree, args, gen_sym, **kw):
+def dletseq(tree, args, *, gen_sym, **kw):
     """[syntax, decorator] Decorator version of letseq, for 'letseq over def'.
 
     Expands to nested function definitions, each with one ``dlet`` decorator.
@@ -293,7 +320,7 @@ def dletseq(tree, args, gen_sym, **kw):
         assert g(10) == 14
     """
     with dyn.let(gen_sym=gen_sym):
-        return _dletseq(bindings=args, fdef=tree)
+        return _destructure_and_apply_let(tree, args, _dletseq)
 
 @macros.decorator
 def dletrec(tree, args, *, gen_sym, **kw):
@@ -311,7 +338,7 @@ def dletrec(tree, args, *, gen_sym, **kw):
     Same cautions apply as to ``dlet``.
     """
     with dyn.let(gen_sym=gen_sym):
-        return _dletrec(bindings=args, fdef=tree)
+        return _destructure_and_apply_let(tree, args, _dletrec)
 
 @macros.decorator
 def blet(tree, args, *, gen_sym, **kw):
@@ -325,10 +352,10 @@ def blet(tree, args, *, gen_sym, **kw):
         assert result == 42
     """
     with dyn.let(gen_sym=gen_sym):
-        return _blet(bindings=args, fdef=tree)
+        return _destructure_and_apply_let(tree, args, _blet)
 
 @macros.decorator
-def bletseq(tree, args, gen_sym, **kw):
+def bletseq(tree, args, *, gen_sym, **kw):
     """[syntax, decorator] def --> letseq block.
 
     Example::
@@ -341,7 +368,7 @@ def bletseq(tree, args, gen_sym, **kw):
         assert result == 4
     """
     with dyn.let(gen_sym=gen_sym):
-        return _bletseq(bindings=args, fdef=tree)
+        return _destructure_and_apply_let(tree, args, _bletseq)
 
 @macros.decorator
 def bletrec(tree, args, *, gen_sym, **kw):
@@ -368,13 +395,13 @@ def bletrec(tree, args, *, gen_sym, **kw):
         assert result is True
     """
     with dyn.let(gen_sym=gen_sym):
-        return _bletrec(bindings=args, fdef=tree)
+        return _destructure_and_apply_let(tree, args, _bletrec)
 
 # -----------------------------------------------------------------------------
 # Imperative code in expression position.
 
 @macros.expr
-def do(tree, gen_sym, **kw):
+def do(tree, *, gen_sym, **kw):
     """[syntax, expr] Stuff imperative code into an expression position.
 
     Return value is the value of the last expression inside the ``do``.
@@ -506,7 +533,7 @@ def do(tree, gen_sym, **kw):
         return _do(tree)
 
 @macros.expr
-def do0(tree, gen_sym, **kw):
+def do0(tree, *, gen_sym, **kw):
     """[syntax, expr] Like do, but return the value of the first expression."""
     with dyn.let(gen_sym=gen_sym):
         return _do0(tree)
@@ -514,7 +541,7 @@ def do0(tree, gen_sym, **kw):
 # -----------------------------------------------------------------------------
 
 @macros.expr
-def let_syntax(tree, args, gen_sym, **kw):
+def let_syntax(tree, args, *, gen_sym, **kw):
     with dyn.let(gen_sym=gen_sym):  # gen_sym is only needed by the implicit do.
         return _destructure_and_apply_let(tree, args, let_syntax_expr, allow_call_in_name_position=True)
 
@@ -629,7 +656,7 @@ def let_syntax(tree, **kw):
     return let_syntax_block(block_body=tree)
 
 @macros.expr
-def abbrev(tree, args, gen_sym, **kw):
+def abbrev(tree, args, *, gen_sym, **kw):
     with dyn.let(gen_sym=gen_sym):  # gen_sym is only needed by the implicit do.
         yield _destructure_and_apply_let(tree, args, let_syntax_expr, allow_call_in_name_position=True)
 
@@ -681,7 +708,7 @@ def forall(tree, **kw):
 # -----------------------------------------------------------------------------
 
 @macros.block
-def multilambda(tree, gen_sym, **kw):
+def multilambda(tree, *, gen_sym, **kw):
     """[syntax, block] Supercharge your lambdas: multiple expressions, local variables.
 
     For all ``lambda`` lexically inside the ``with multilambda`` block,
@@ -717,38 +744,44 @@ def multilambda(tree, gen_sym, **kw):
 
 @macros.block
 def namedlambda(tree, **kw):
-    """[syntax, block] Implicitly named lambdas.
+    """[syntax, block] Name lambdas implicitly.
 
     Lexically inside a ``with namedlambda`` block, any literal ``lambda``
-    that is assigned to a name using a simple assignment of the form
-    ``f = lambda ...: ...``, is named as "f", where the name ``f``
-    is captured from the assignment statement at macro expansion time.
+    that is assigned to a name using one of the supported assignment forms
+    is named to have the name of the LHS of the assignment. The name is
+    captured at macro expansion time.
 
-    For capturing the name, the assignment must be of a single ``lambda`` value
-    to a single name; other forms of assignment are not supported. (This may be
-    subject to change in a future version.)
+    Naming modifies the original function object.
 
-    Additionally, during the dynamic extent of the ``with namedlambda`` block,
-    assigning a lambda to a name in an ``unpythonic.env`` instance will cause
-    that lambda to be named, capturing the name it is assigned to in the env.
-    This is performed at run time.
+    We support:
 
-    Naming modifies the original function object (specifically, its ``__name__``
-    and ``__qualname__`` attributes). The name is set only once per object, so in::
+        - Single-item assignments to a local name, ``f = lambda ...: ...``
+
+        - Assignments to unpythonic environments, ``f << (lambda ...: ...)``
+
+        - Let bindings, ``let[(f, (lambda ...: ...)) in ...]``, using any
+          let syntax supported by unpythonic (here using the haskelly let-in
+          just as an example).
+
+    Support for other forms of assignment might or might not be added in a
+    future version.
+
+    Example::
 
         with namedlambda:
-            f = lambda x: x**3        # lexical rule: name as "f"
+            f = lambda x: x**3        # assignment: name as "f"
 
             let((x, 42), (g, None), (h, None))[[
-              g << (lambda x: x**2),  # dynamic rule: name as "g"
-              h << f,                 # no-rename rule: still "f"
+              g << (lambda x: x**2),  # env-assignment: name as "g"
+              h << f,                 # still "f" (no literal lambda on RHS)
               (g(x), h(x))]]
 
-    the name of the first lambda will be set as ``f``, and it will remain as ``f``
-    even after the name ``h`` is made to point to the same object inside the
-    body of the ``let``.
+            foo = let[(f7, lambda x: x) in f7]  # let-binding: name as "f7"
+
+    The naming is performed using the function ``unpythonic.misc.namelambda``,
+    which will update ``__name__``, ``__qualname__`` and ``__code__.co_name``.
     """
-    return _namedlambda(block_body=tree)
+    return (yield from _namedlambda(block_body=tree))
 
 @macros.block
 def quicklambda(tree, **kw):
@@ -783,32 +816,29 @@ def quicklambda(tree, **kw):
     """
     return (yield from _quicklambda(block_body=tree))
 
-# -----------------------------------------------------------------------------
+@macros.block
+def envify(tree, *, gen_sym, **kw):
+    """[syntax, block] Make formal parameters live in an unpythonic env.
 
-@macros.expr
-def fup(tree, **kw):
-    """[syntax, expr] Functionally update a sequence.
+    The purpose is to allow overwriting formals using unpythonic's
+    expression-assignment ``name << value``. The price is that the references
+    to the arguments are copied into an env whenever an envified function is
+    entered.
 
-    Example::
+    Example - PG's accumulator puzzle (http://paulgraham.com/icad.html)::
 
-        from itertools import repeat
+        with envify:
+            def foo(n):
+                return lambda i: n << n + i
 
-        lst = (1, 2, 3, 4, 5)
-        assert fup[lst[3] << 42] == (1, 2, 3, 42, 5)
-        assert fup[lst[0::2] << tuple(repeat(10, 3))] == (10, 2, 10, 4, 10)
+    Of course, now we can::
 
-    The transformation is::
-
-        fup[seq[idx] << value] --> fupdate(seq, idx, value)
-        fup[seq[slicestx] << iterable] --> fupdate(seq, slice(...), iterable)
-
-    Limitations:
-
-      - Currently only one update specification is supported in a single ``fup[]``.
-
-    Named after the sound a sequence makes when it is hit by a functional update.
+        with autoreturn, envify:
+            def foo(n):
+                lambda i: n << n + i
     """
-    return _fup(tree)
+    with dyn.let(gen_sym=gen_sym):
+        return (yield from _envify(block_body=tree))
 
 # -----------------------------------------------------------------------------
 
@@ -973,7 +1003,7 @@ def tco(tree, *, gen_sym, **kw):
         return (yield from _tco(block_body=tree))
 
 @macros.block
-def continuations(tree, gen_sym, **kw):
+def continuations(tree, *, gen_sym, **kw):
     """[syntax, block] call/cc for Python.
 
     This allows saving the control state and then jumping back later
@@ -1372,6 +1402,408 @@ def continuations(tree, gen_sym, **kw):
 
 # -----------------------------------------------------------------------------
 
+@macros.block
+def nb(tree, args, **kw):
+    """[syntax, block] Ultralight math notebook.
+
+    Auto-print top-level expressions, auto-assign last result as _.
+
+    A custom print function can be supplied as the first positional argument.
+
+    Example::
+
+        with nb:
+            2 + 3
+            42 * _
+
+        from sympy import *
+        with nb(pprint):
+            x, y = symbols("x, y")
+            x * y
+            3 * _
+    """
+    return _nb(body=tree, args=args)
+
+# -----------------------------------------------------------------------------
+
+@macros.block
+def lazify(tree, *, gen_sym, **kw):
+    """[syntax, block] Call-by-need for Python.
+
+    In a ``with lazify`` block, function arguments are evaluated only when
+    actually used, at most once each, and in the order in which they are
+    actually used. Promises are automatically forced on access.
+
+    Automatic lazification applies to arguments in function calls and to
+    let-bindings, since they play a similar role. **No other binding forms
+    are auto-lazified.**
+
+    Automatic lazification uses the ``lazyrec[]`` macro, which recurses into
+    certain types of container literals, so that the lazification will not
+    interfere with unpacking. See its docstring for details.
+
+    Comboing with other block macros in ``unpythonic.syntax`` is supported,
+    including ``curry`` and ``continuations``.
+
+    Silly contrived example::
+
+        with lazify:
+            def my_if(p, a, b):
+                if p:
+                    return a  # b never evaluated in this code path...
+                else:
+                    return b  # a never evaluated in this code path...
+
+            # ...hence the divisions by zero here are never performed.
+            assert my_if(True, 23, 1/0) == 23
+            assert my_if(False, 1/0, 42) == 42
+
+    Note ``my_if`` is a run-of-the-mill runtime function, not a macro. Only the
+    ``with lazify`` is imbued with any magic.
+
+    Like ``with continuations``, no state or context is associated with a
+    ``with lazify`` block, so lazy functions defined in one block may call
+    those defined in another. Calls between lazy and strict code are also
+    supported (in both directions), without requiring any extra effort.
+
+    Evaluation of each lazified argument is guaranteed to occur at most once;
+    the value is cached. Order of evaluation of lazy arguments is determined
+    by the (dynamic) order in which the lazy code actually uses them.
+
+    Essentially, the above code expands into::
+
+        from macropy.quick_lambda import macros, lazy
+        from unpythonic.syntax import force
+
+        def my_if(p, a, b):
+            if force(p):
+                return force(a)
+            else:
+                return force(b)
+        assert my_if(lazy[True], lazy[23], lazy[1/0]) == 23
+        assert my_if(lazy[False], lazy[1/0], lazy[42]) == 42
+
+    plus some clerical details to allow lazy and strict code to be mixed.
+
+    Just passing through a lazy argument to another lazy function will
+    not trigger evaluation, even when it appears in a computation inlined
+    to the argument list::
+
+        with lazify:
+            def g(a, b):
+                return a
+            def f(a, b):
+                return g(2*a, 3*b)
+            assert f(21, 1/0) == 42
+
+    The division by zero is never performed, because the value of ``b`` is
+    not needed to compute the result (worded less magically, that promise is
+    never forced in the code path that produces the result). Essentially,
+    the above code expands into::
+
+        from macropy.quick_lambda import macros, lazy
+        from unpythonic.syntax import force
+
+        def g(a, b):
+            return force(a)
+        def f(a, b):
+            return g(lazy[2*force(a)], lazy[3*force(b)])
+        assert f(lazy[21], lazy[1/0]) == 42
+
+    This relies on the magic of closures to capture f's ``a`` and ``b`` into
+    the promises.
+
+    But be careful; **assignments are not auto-lazified**, so the following does
+    **not** work::
+
+        with lazify:
+            def g(a, b):
+                return a
+            def f(a, b):
+                c = 3*b  # not in an arglist, b gets evaluated!
+                return g(2*a, c)
+            assert f(21, 1/0) == 42
+
+    To avoid that, explicitly wrap the computation into a ``lazy[]``. For why
+    assignment RHSs are not auto-lazified, see the section on pitfalls below.
+
+    In calls, bare references (name, subscript, attribute) are detected and for
+    them, re-thunking is skipped. For example::
+
+        def g(a):
+            return a
+        def f(a):
+            return g(a)
+        assert f(42) == 42
+
+    expands into::
+
+        def g(a):
+            return force(a)
+        def f(a):
+            return g(a)  # <-- no lazy[force(a)] since "a" is just a name
+        assert f(lazy[42]) == 42
+
+    When resolving references, subscripts and attributes are forced just enough
+    to obtain the containing object from a promise, if any; for example, the
+    elements of a list ``lst`` will not be evaluated just because the user code
+    happens to use ``lst.append(...)``; this only forces the object ``lst``
+    itself.
+
+    A ``lst`` appearing by itself evaluates the whole list. Similarly, ``lst[0]``
+    by itself evaluates only the first element, and ``lst[:-1]`` by itself
+    evaluates all but the last element. The index expression in a subscript is
+    fully forced, because its value is needed to determine which elements of the
+    subscripted container are to be accessed.
+
+    **Mixing lazy and strict code**
+
+    Lazy code is allowed to call strict functions and vice versa, without
+    requiring any additional effort.
+
+    Keep in mind what this implies: when calling a strict function, any arguments
+    given to it will be evaluated!
+
+    In the other direction, when calling a lazy function from strict code, the
+    arguments are evaluated by the caller before the lazy code gets control.
+    The lazy code gets just the evaluated values.
+
+    If you have, in strict code, an argument expression you want to pass lazily,
+    use syntax like ``f(lazy[...], ...)``. If you accidentally do this in lazy
+    code, it shouldn't break anything; ``with lazify`` detects any argument
+    expressions that are already promises, and just passes them through.
+
+    **Forcing promises manually**
+
+    This is mainly useful if you ``lazy[]`` or ``lazyrec[]`` something explicitly,
+    and want to compute its value outside a ``with lazify`` block.
+
+    We provide the functions ``force1`` and ``force``.
+
+    Using ``force1``, if ``x`` is a MacroPy ``lazy[]`` promise, it will be
+    forced, and the resulting value is returned. If ``x`` is not a promise,
+    ``x`` itself is returned, à la Racket.
+
+    The function ``force``, in addition, descends into containers (recursively).
+    When an atom ``x`` (i.e. anything that is not a container) is encountered,
+    it is processed using ``force1``.
+
+    Mutable containers are updated in-place; for immutables, a new instance is
+    created. Any container with a compatible ``collections.abc`` is supported.
+    (See ``unpythonic.collections.mogrify`` for details.) In addition, as
+    special cases ``unpythonic.collections.box`` and ``unpythonic.llist.cons``
+    are supported.
+
+    **Tips, tricks and pitfalls**
+
+    You can mix and match bare data values and promises, since ``force(x)``
+    evaluates to ``x`` when ``x`` is not a promise.
+
+    So this is just fine::
+
+        with lazify:
+            def f(x):
+                x = 2*21  # assign a bare data value
+                print(x)  # the implicit force(x) evaluates to x
+            f(17)
+
+    If you want to manually introduce a promise, use ``lazy[]`` from MacroPy::
+
+        from macropy.quick_lambda import macros, lazy
+        from unpythonic.syntax import macros, lazify
+
+        with lazify:
+            def f(x):
+                x = lazy[2*21]  # assign a promise
+                print(x)        # the implicit force(x) evaluates the promise
+            f(17)
+
+    If you have a container literal and want to lazify it recursively in a
+    position that does not auto-lazify, use ``lazyrec[]`` (see its docstring
+    for details)::
+
+        from unpythonic.syntax import macros, lazify, lazyrec
+
+        with lazify:
+            def f(x):
+                return x[:-1]
+            lst = lazyrec[[1, 2, 3/0]]
+            assert f(lst) == [1, 2]
+
+    For non-literal containers, use ``lazy[]`` for each item as appropriate::
+
+        def f(lst):
+            lst.append(lazy["I'm lazy"])
+            lst.append(lazy["Don't call me lazy, I'm just evaluated later!"])
+
+    Keep in mind, though, that ``lazy[]`` will introduce a lambda, so there's
+    the usual pitfall::
+
+        from macropy.quick_lambda import macros, lazy
+        from unpythonic.syntax import macros, lazify
+
+        with lazify:
+            lst = []
+            for x in range(3):       # DANGER: only one "x", mutated imperatively
+                lst.append(lazy[x])  # all these closures capture the same "x"
+            print(lst[0])  # 2
+            print(lst[1])  # 2
+            print(lst[2])  # 2
+
+    So to capture the value instead of the name, use the usual workaround,
+    the wrapper lambda (here written more readably as a let, which it really is)::
+
+        from macropy.quick_lambda import macros, lazy
+        from unpythonic.syntax import macros, lazify, let
+
+        with lazify:
+            lst = []
+            for x in range(3):
+                lst.append(let[(y, x) in lazy[y]])
+            print(lst[0])  # 0
+            print(lst[1])  # 1
+            print(lst[2])  # 2
+
+    Be careful not to ``lazy[]`` or ``lazyrec[]`` too much::
+
+        with lazify:
+            a = 10
+            a = lazy[2*a]  # 20, right?
+            print(a)       # crash!
+
+    Why does this example crash? The expanded code is::
+
+        with lazify:
+            a = 10
+            a = lazy[2*force(a)]
+            print(force(a))
+
+    The ``lazy[]`` sets up a promise, which will force ``a`` *at the time when
+    the containing promise is forced*, but at that time the name ``a`` points
+    to a promise, which will force...
+
+    The fundamental issue is that ``a = 2*a`` is an imperative update; if you
+    need to do that, just let Python evaluate the RHS normally (i.e. use the
+    value the name ``a`` points to *at the time when the RHS runs*).
+
+    Assigning a lazy value to a new name evaluates it, because any read access
+    triggers evaluation::
+
+        with lazify:
+            def g(x):
+                y = x       # the "x" on the RHS triggers the implicit force
+                print(y)    # bare data value
+            f(2*21)
+
+    Inspired by Haskell, Racket's (delay) and (force), and lazy/racket.
+
+    **Combos**
+
+    Introducing the *HasThon* programming language (it has 100% more Thon than
+    popular brands)::
+
+        with curry, lazify:  # or continuations, curry, lazify if you want those
+            def add2first(a, b, c):
+                return a + b
+            assert add2first(2)(3)(1/0) == 5
+
+            def f(a, b):
+                return a
+            assert let[((c, 42),
+                        (d, 1/0)) in f(c)(d)] == 42
+            assert letrec[((c, 42),
+                           (d, 1/0),
+                           (e, 2*c)) in f(e)(d)] == 84
+
+            assert letrec[((c, 42),
+                           (d, 1/0),
+                           (e, 2*c)) in [local[x << f(e)(d)],
+                                         x/4]] == 21
+
+    Works also with continuations. Rules:
+
+      - Also continuations are transformed into lazy functions.
+
+      - ``cc`` built by chain_conts is treated as lazy, **itself**; then it's
+        up to the continuations chained by it to decide whether to force their
+        arguments.
+
+      - The default continuation ``identity`` is strict, so that return values
+        from a continuation-enabled computation will be forced.
+
+    Example::
+
+        with continuations, lazify:
+            k = None
+            def setk(*args, cc):
+                nonlocal k
+                k = cc
+                return args[0]
+            def doit():
+                lst = ['the call returned']
+                *more, = call_cc[setk('A', 1/0)]
+                return lst + [more[0]]
+            assert doit() == ['the call returned', 'A']
+            assert k('again') == ['the call returned', 'again']
+            assert k('thrice', 1/0) == ['the call returned', 'thrice']
+
+    For a version with comments, see ``unpythonic/syntax/test/test_lazify.py``.
+
+    **CAUTION**: Call-by-need is a low-level language feature that is difficult
+    to bolt on after the fact. Some things might not work.
+
+    **CAUTION**: The functions in ``unpythonic.fun`` are lazify-aware (so that
+    e.g. curry and compose work with lazy functions), as are ``call`` and
+    ``callwith`` in ``unpythonic.misc``, but the rest of ``unpythonic`` is not.
+
+    **CAUTION**: Argument passing by function call, and let-bindings are
+    currently the only binding constructs to which auto-lazification is applied.
+    """
+    with dyn.let(gen_sym=gen_sym):
+        return (yield from _lazify(body=tree))
+
+@macros.expr
+def lazyrec(tree, **kw):
+    """[syntax, expr] Delay items in a container literal, recursively.
+
+    Essentially, this distributes ``lazy[]`` into the items inside a literal
+    ``list``, ``tuple``, ``set``, ``frozenset``, ``unpythonic.collections.box``
+    or ``unpythonic.llist.cons``, and into the values of a literal ``dict`` or
+    ``unpythonic.collections.frozendict``.
+
+    Because this is a macro and must work by names only, only this fixed set of
+    container types is supported.
+
+    The container itself is not lazified, only the items inside it are, to keep
+    the lazification from interfering with unpacking. This allows things such as
+    ``f(*lazyrec[(1*2*3, 4*5*6)])`` to work as expected.
+
+    See also ``macropy.quick_lambda.lazy`` (the effect on each item) and
+    ``unpythonic.syntax.force`` (the inverse of ``lazyrec[]``).
+
+    For an atom, ``lazyrec[]`` has the same effect as ``lazy[]``::
+
+        lazyrec[dostuff()] --> lazy[dostuff()]
+
+    For a container literal, ``lazyrec[]`` descends into it::
+
+        lazyrec[(2*21, 1/0)] --> (lazy[2*21], lazy[1/0])
+        lazyrec[{'a': 2*21, 'b': 1/0}] --> {'a': lazy[2*21], 'b': lazy[1/0]}
+
+    Constructor call syntax for container literals is also supported::
+
+        lazyrec[list(2*21, 1/0)] --> [lazy[2*21], lazy[1/0]]
+
+    Nested container literals (with any combination of known types) are
+    processed recursively, for example::
+
+        lazyrec[((2*21, 1/0), (1+2+3, 4+5+6))] --> ((lazy[2*21], lazy[1/0]),
+                                                    (lazy[1+2+3], lazy[4+5+6]))
+    """
+    return _lazyrec(tree)
+
+# -----------------------------------------------------------------------------
+
 @macros.block
 def prefix(tree, **kw):
     """[syntax, block] Write Python like Lisp: the first item is the operator.
@@ -1442,6 +1874,8 @@ def prefix(tree, **kw):
 
         - For ``*args``, to keep it lispy, maybe you want ``unpythonic.fun.apply``;
           this allows syntax such as ``(apply, f, 1, 2, lst)``.
+
+    **CAUTION**: This macro is experimental, not intended for production use.
     """
     return (yield from _prefix(block_body=tree))
 
