@@ -20,18 +20,18 @@ from macropy.core.quotes import macros, q, u, ast_literal, name
 from macropy.core.hquotes import macros, hq
 from macropy.core.walkers import Walker
 
-from .util import isx, isec, \
+from .util import isx, make_isxpred, isec, \
                   detect_callec, detect_lambda, \
                   has_tco, sort_lambda_decorators, \
-                  suggest_decorator_index
-from .letdoutil import isdo, islet
+                  suggest_decorator_index, ContinuationsMarker, wrapwith, ismarker
+from .letdoutil import isdo, islet, ExpandedLetView, ExpandedDoView
 from .ifexprs import aif
-from .letdo import let
 
 from ..dynassign import dyn
 from ..it import uniqify
-from ..fun import identity
+from ..fun import identity, orf
 from ..tco import trampolined, jump
+from ..lazyutil import mark_lazy
 
 # -----------------------------------------------------------------------------
 # Implicit return. This performs a tail-position analysis of function bodies.
@@ -72,13 +72,19 @@ def autoreturn(block_body):
 def tco(block_body):
     # first pass, outside-in
     userlambdas = detect_lambda.collect(block_body)
-    known_ecs = list(uniqify(detect_callec.collect(block_body)))
+    known_ecs = list(uniqify(detect_callec(block_body)))
     block_body = yield block_body
 
     # second pass, inside-out
     transform_retexpr = partial(_transform_retexpr)
     new_block_body = []
     for stmt in block_body:
+        # skip nested, already expanded "with continuations" blocks
+        # (needed to support continuations in the Lispython dialect, which applies tco globally)
+        if ismarker("ContinuationsMarker", stmt):
+            new_block_body.append(stmt)
+            continue
+
         stmt = _tco_transform_return.recurse(stmt, known_ecs=known_ecs,
                                              transform_retexpr=transform_retexpr)
         stmt = _tco_transform_def.recurse(stmt, preproc_cb=None)
@@ -118,6 +124,42 @@ def call_cc(tree, **kw):
     """
     pass
 
+# _pcc/cc chaining handler, to be exported to client code via hq[].
+#
+# We handle multiple-return-values like the rest of unpythonic does:
+# returning a tuple means returning multiple values. Unpack them
+# to cc's arglist.
+#
+def chain_conts(cc1, cc2, with_star=False):  # cc1=_pcc, cc2=cc
+    """Internal function, used in code generated by the continuations macro."""
+    if with_star:  # to be chainable from a tail call, accept a multiple-values arglist
+        if cc1 is not None:
+            @mark_lazy
+            def cc(*value):
+                return jump(cc1, cc=cc2, *value)
+        else:
+            # Beside a small optimization, it is important to preserve
+            # "identity" as "identity", so that the call_cc logic that
+            # defines the continuation functions will detect it and
+            # know when to set _pcc (and importantly, when not to).
+            cc = cc2
+    else:  # for inert data value returns (this produces the multiple-values arglist)
+        if cc1 is not None:
+            @mark_lazy
+            def cc(value):
+                if isinstance(value, tuple):
+                    return jump(cc1, cc=cc2, *value)
+                else:
+                    return jump(cc1, value, cc=cc2)
+        else:
+            @mark_lazy
+            def cc(value):
+                if isinstance(value, tuple):
+                    return jump(cc2, *value)
+                else:
+                    return jump(cc2, value)
+    return cc
+
 def continuations(block_body):
     # This is a very loose pythonification of Paul Graham's continuation-passing
     # macros in On Lisp, chapter 20.
@@ -127,7 +169,7 @@ def continuations(block_body):
 
     # first pass, outside-in
     userlambdas = detect_lambda.collect(block_body)
-    known_ecs = list(uniqify(detect_callec.collect(block_body)))
+    known_ecs = list(uniqify(detect_callec(block_body)))
     block_body = yield block_body
 
     # second pass, inside-out
@@ -160,7 +202,7 @@ def continuations(block_body):
                 tree.args.kw_defaults[j] = hq[identity]
         # implicitly add "parent cc" arg for treating the tail of a computation
         # as one entity (only actually used in continuation definitions created by
-        # call_cc; everywhere else, it's None). See the PDF for clarifying pictures.
+        # call_cc; everywhere else, it's None). See callcc_topology.pdf for clarifying pictures.
         if "_pcc" not in kwonlynames:
             non = q[None]
             non = copy_location(non, tree)
@@ -193,13 +235,13 @@ def continuations(block_body):
             # chain our _pcc and the cc=... manually provided by the user
             thekw = [kw for kw in tree.keywords if kw.arg == "cc"][0]  # exactly one
             usercc = thekw.value
-            thekw.value = q[chain_conts(name["_pcc"], ast_literal[usercc], with_star=True)]
+            thekw.value = hq[chain_conts(name["_pcc"], ast_literal[usercc], with_star=True)]
         else:
             # chain our _pcc and the current value of cc
-            tree.keywords = [keyword(arg="cc", value=q[chain_conts(name["_pcc"], name["cc"], with_star=True)])] + tree.keywords
+            tree.keywords = [keyword(arg="cc", value=hq[chain_conts(name["_pcc"], name["cc"], with_star=True)])] + tree.keywords
         return tree
     def data_cb(tree):  # transform an inert-data return value into a tail-call to cc.
-        tree = q[chain_conts(name["_pcc"], name["cc"])(ast_literal[tree])]
+        tree = hq[chain_conts(name["_pcc"], name["cc"])(ast_literal[tree])]
         return tree
     transform_retexpr = partial(_transform_retexpr, call_cb=call_cb, data_cb=data_cb)
 
@@ -227,6 +269,9 @@ def continuations(block_body):
             stmt, *after = after
             if iscallcc(stmt):
                 # after is always non-empty here (has at least the explicitified "return")
+                # ...unless we're at the top level of the "with continuations" block
+                if not after:
+                    assert False, "call_cc[] cannot appear as the last statement of a 'with continuations' block (no continuation to capture)"
                 return before, stmt, after
             before.append(stmt)
             if not after:
@@ -347,7 +392,8 @@ def continuations(block_body):
                                       kw_defaults=[hq[identity], maybe_capture]),
                        body=contbody,
                        decorator_list=[],  # patched later by transform_def
-                       returns=None)  # return annotation not used here
+                       returns=None,  # return annotation not used here
+                       lineno=locref.lineno, col_offset=locref.col_offset)
 
         # in the output stmts, define the continuation function...
         newstmts = [funcdef]
@@ -436,39 +482,6 @@ def continuations(block_body):
     # (the top-level "cc" is only used for continuations created by call_cc[] at the top level of the block)
     new_block_body = [Assign(targets=[q[name["cc"]]], value=hq[identity])]
 
-    # Define the _pcc/cc chaining handler (cleaner generated code if we centralize this)
-    #
-    # Handle multiple-return-values like the rest of unpythonic does:
-    # returning a tuple means returning multiple values. Unpack them
-    # to cc's arglist.
-    with hq as chain_conts:
-        def chain_conts(cc1, cc2, with_star=False):  # cc1=_pcc, cc2=cc
-            if with_star:  # to be chainable from a tail call, accept a multiple-values arglist
-                if cc1 is not None:
-                    def cc(*value):
-                        return jump(cc1, cc=cc2, *value)
-                else:
-                    # Beside a small optimization, it is important to preserve
-                    # "identity" as "identity", so that the call_cc logic that
-                    # defines the continuation functions will detect it and
-                    # know when to set _pcc (and importantly, when not to).
-                    cc = cc2
-            else:  # for inert data value returns (this produces the multiple-values arglist)
-                if cc1 is not None:
-                    def cc(value):
-                        if isinstance(value, tuple):
-                            return jump(cc1, cc=cc2, *value)
-                        else:
-                            return jump(cc1, value, cc=cc2)
-                else:
-                    def cc(value):
-                        if isinstance(value, tuple):
-                            return jump(cc2, *value)
-                        else:
-                            return jump(cc2, value)
-            return cc
-    new_block_body.append(chain_conts)
-
     # transform all defs (except the chaining handler), including those added by call_cc[].
     for stmt in block_body:
         stmt = _tco_transform_def.recurse(stmt, preproc_cb=transform_args)
@@ -478,7 +491,12 @@ def continuations(block_body):
                                              transform_retexpr=transform_retexpr)
         stmt = sort_lambda_decorators(stmt)
         new_block_body.append(stmt)
-    return new_block_body
+
+    # Leave a marker so "with tco", if applied, can ignore the expanded "with continuations" block
+    # (needed to support continuations in the Lispython dialect, since it applies tco globally.)
+    return wrapwith(item=hq[ContinuationsMarker],
+                    body=new_block_body,
+                    locref=block_body[0])
 
 # -----------------------------------------------------------------------------
 
@@ -550,6 +568,7 @@ def _tco_transform_lambda(tree, *, preproc_cb, userlambdas, known_ecs, transform
 
 # Tail-position analysis for a return-value expression (also the body of a lambda).
 # Here we need to be very, very selective about where to recurse so this is not a Walker.
+_isjump = orf(make_isxpred("jump"), make_isxpred("loop"))
 def _transform_retexpr(tree, known_ecs, call_cb=None, data_cb=None):
     """Analyze and TCO a return-value expression or a lambda body.
 
@@ -573,14 +592,22 @@ def _transform_retexpr(tree, known_ecs, call_cb=None, data_cb=None):
     transform_call = call_cb or (lambda tree: tree)
     transform_data = data_cb or (lambda tree: tree)
     def transform(tree):
-        if isdo(tree) or islet(tree):
-            # Ignore the "lambda e: ...", and descend into the ..., in:
-            #   - let[] or letrec[] in tail position.
-            #     - letseq[] is a nested sequence of lets, so covers that too.
-            #   - do[] in tail position.
-            #     - May be generated also by a "with multilambda" block
-            #       that has already expanded.
-            tree.args[-1].body = transform(tree.args[-1].body)
+        # Ignore the "lambda e: ...", and descend into the ..., in:
+        #   - let[] or letrec[] in tail position.
+        #     - letseq[] is a nested sequence of lets, so covers that too.
+        #   - do[] in tail position.
+        #     - May be generated also by a "with multilambda" block
+        #       that has already expanded.
+        if islet(tree):
+            view = ExpandedLetView(tree)
+            assert view.body, "BUG: what's this, a decorator inside a lambda?"
+            thelambda = view.body  # lambda e: ...
+            thelambda.body = transform(thelambda.body)
+        elif isdo(tree):
+            thebody = ExpandedDoView(tree).body   # list of do-items
+            lastitem = thebody[-1]  # lambda e: ...
+            thelambda = lastitem
+            thelambda.body = transform(thelambda.body)
         elif type(tree) is Call:
             # Apply TCO to tail calls.
             #   - If already an explicit jump() or loop(), leave it alone.
@@ -592,7 +619,7 @@ def _transform_retexpr(tree, known_ecs, call_cb=None, data_cb=None):
             #     - Hence, transform_return() calls us on the content of
             #       all ec nodes directly. ec(...) is like return; the
             #       argument is the retexpr.
-            if not (isx(tree.func, "jump") or isx(tree.func, "loop") or isec(tree, known_ecs)):
+            if not (isx(tree.func, _isjump) or isec(tree, known_ecs)):
                 tree.args = [tree.func] + tree.args
                 tree.func = hq[jump]
                 tree = transform_call(tree)
